import unittest
import os
import filecmp
import subprocess


EXEC_NAME = './myunzip0'

class TestMyUnzip0(unittest.TestCase):

    def test_uncompressed(self):
        subprocess.run([EXEC_NAME, 'tests/resources/uncompressed.zip'])
        
        # Check that the correct file was extracted
        self.assertTrue(os.path.exists('tests/resources/uncompressed.txt'))

        # Check correct output 
        self.assertTrue(filecmp.cmp('tests/resources/uncompressed.txt', 
                                    'tests/resources/uncompressed_expected.txt'))

        # Remove output 
        os.remove('tests/resources/uncompressed.txt')

    def test_compressed(self):
        subprocess.run([EXEC_NAME, 'tests/resources/compressed.zip'])

        # Check that we extracted a DEFLATE stream 
        self.assertTrue(os.path.exists('tests/resources/compressed.txt.deflate'))

        # Check correct output
        self.assertTrue(filecmp.cmp('tests/resources/compressed_expected.deflate', 
                                    'tests/resources/compressed.txt.deflate'))

        # Remove output 
        os.remove('tests/resources/compressed.txt.deflate')

if __name__ == '__main__':
    unittest.main()
